#!/usr/bin/env python3
"""
Basic configuration of 2DOF motion simulator motors

Build thread:
https://www.xsimulator.net/community/threads/2dof-diy-simulator-brushless-by-cliojohn.13753/

Motor Type:
    Brushless motors: Dualsky 5330C - v2

Encoder Type:
    AMT102-V optical encoder

Odrive Controller board type:
    ODrive v3.5 - 56 Volt

Odrive Firmware version:
    0.5.4 v1

End position switches:
    In use, normally closed, "config style 3" --> active high, 10k pull-up to 5V
    Using GPIOs #1 (max) and #2 (min) for Axis 0 and 
    Using GPIOs #3 (max) and #4 (min) for Axis 1 
*******************************************************************************************
Status/script version log:
06.09.22 - Added startup configuration lines to be set after initial tuning.
03.09.22 - Tested with two separate motors unloaded while configuring and then loaded.
01.09.22 - Commented out axis1 main method until ready for test

*******************************************************************************************
These configurations is based on Austin Owens implementation:
https://github.com/AustinOwens/robodog/blob/main/odrive/configs/odrive_hoverboard_config.py
*******************************************************************************************
@cliojohn, jt.andreassen@gmail.com

"""

import sys
import time
import odrive
import fibre # required to handle USB comm errors etc.
from odrive.enums import *

class BrushlessMotorConfig:
    """
    Class for configuring an Odrive axis for a Brushless 225kV motor.
    Only works with one Odrive at a time.
    """
    MOTOR_KV_VALUE = 225

    # Min/Max phase inductance of motor
    MIN_PHASE_INDUCTANCE = 0
    MAX_PHASE_INDUCTANCE = 0.002

    # Min/Max phase resistance of motor
    MIN_PHASE_RESISTANCE = 0
    MAX_PHASE_RESISTANCE = 1  # Nominal 0.042

    # Tolerance for encoder offset float
    ENCODER_OFFSET_FLOAT_TOLERANCE = 0.5

    def __init__(self, axis_num: int):
        """
        Initalizes BrushlessMotorConfig class by finding odrive and grabbing specified
        axis object.

        :param axis_num: Which channel/motor on the odrive your referring to.
        :type axis_num: int (0 or 1)
        """
        self.axis_num = axis_num
        
        # Connect to Odrive
        print("Looking for ODrive...")
        self._find_odrive()
        print("Found ODrive.")

    def _find_odrive(self):
        # connect to Odrive
        self.odrv = odrive.find_any()
        self.odrv_axis = getattr(self.odrv, "axis{}".format(self.axis_num))
    #   self.odrv.channel._channel_broken.subscribe(self.USBerror)
    #    fibre.shell.did_discover_device(self.odrv,self.log,self.token1)


    #def USBerror():
    #    print("Device disconnected")

    def erase_config(self):
        """ Erase pre-exsisting configuration """

        print("Erasing pre-exsisting configuration...")
        try:
            self.odrv.erase_configuration()
        except fibre.libfibre.ObjectLostError: # The Odrive will reboot
            print("Device disconnected")
            pass

    def set_odrive_parameters(self):
        self._find_odrive() # Odrive erased + rebooted? = Reconnect.

        """Saves odrive axis, motor, encoder and controller parameters"""
        self.odrv_axis.motor.config.pole_pairs = 7
        self.odrv_axis.motor.config.motor_type = MOTOR_TYPE_HIGH_CURRENT

        self.odrv.config.enable_brake_resistor = True  # Default resistor
        self.odrv.config.brake_resistance = 2          # Default = 2

        self.odrv_axis.motor.config.resistance_calib_max_voltage = 2
        self.odrv_axis.motor.config.requested_current_range      = 60
        self.odrv_axis.motor.config.current_control_bandwidth    = 1000

        self.odrv_axis.motor.config.torque_constant = 8.27/self.MOTOR_KV_VALUE  # Around 0.04 for 225kV

        # We are going to use AMT102-V encoder.
        self.odrv_axis.encoder.config.mode=ENCODER_MODE_INCREMENTAL  # = 0
        self.odrv_axis.encoder.config.cpr=8192   # 4 x Pulse Per Revolution setting = 4 x 2048
        self.odrv_axis.encoder.config.use_index = True
        self.odrv_axis.config.startup_encoder_index_search = True

        # Set params for start-up:
        #self.odrv_axis.config.startup_motor_calibration
        #self.odrv_axis.config.startup_encoder_index_search
        #self.odrv_axis.config.startup_encoder_offset_calibration
        #self.odrv_axis.config.startup_closed_loop_control
        
        # Set params for end position switches:
        self.odrv.config.gpio1_mode = GPIO_MODE_DIGITAL       # = 0 
        self.odrv.config.gpio2_mode = GPIO_MODE_DIGITAL
        self.odrv.config.gpio3_mode = GPIO_MODE_DIGITAL
        self.odrv.config.gpio4_mode = GPIO_MODE_DIGITAL

        self.odrv.config.gpio1_mode = GPIO_MODE_DIGITAL_PULL_UP # = 1 --> Config style 3 (NC)
        self.odrv.config.gpio2_mode = GPIO_MODE_DIGITAL_PULL_UP
        self.odrv.config.gpio3_mode = GPIO_MODE_DIGITAL_PULL_UP
        self.odrv.config.gpio4_mode = GPIO_MODE_DIGITAL_PULL_UP

        # Allocate GPIOs to correct axis 0 or 1
        if self.axis_num == 0:
            self.odrv_axis.max_endstop.config.gpio_num = 1
            self.odrv_axis.min_endstop.config.gpio_num = 2
        if self.axis_num == 1:
            self.odrv_axis.max_endstop.config.gpio_num = 3
            self.odrv_axis.min_endstop.config.gpio_num = 4

        # Set correct config style for end pos. switches + enable
        self.odrv_axis.min_endstop.config.offset = -0.25        # Move N turns from minimum
        self.odrv_axis.min_endstop.config.is_active_high = True
        self.odrv_axis.max_endstop.config.is_active_high = True
        self.odrv_axis.min_endstop.config.enabled = True
        self.odrv_axis.max_endstop.config.enabled = True

        # TODO: Tuned values
        self.odrv_axis.controller.config.pos_gain = 72  # Original is 20 and that is slow on 225kV motors
        self.odrv_axis.controller.config.vel_gain = 0.25 # Original is 0.166
        self.odrv_axis.controller.config.vel_integrator_gain = 1.19 # Original is 0.33, tuning val = 1.19 (HOLD/trips?)
        self.odrv_axis.controller.config.vel_limit = 10000 # TODO: set to between 50000 and 120000 propably at the end
        # TODO: motor current limit to around 30 Amps (max = "axis.config.requested_current_range") after tuning
        #self.odrv_axis.motor.config.current_lim = 30   # Default at 10A for tuning   

        # Set in position control mode so we can control the position with input angles
        self.odrv_axis.controller.config.control_mode = CONTROL_MODE_POSITION_CONTROL

        # In the next step we are going to start powering the motor and so we
        # want to make sure that some of the above settings that require a
        # reboot are applied first.
        print("Saving manual configuration and rebooting...")
        try:
            self.odrv.save_configuration()
            print("Manual configuration saved.")

        except fibre.libfibre.ObjectLostError:
            print("Device disconnected")
            pass # Saving configuration makes the device reboot

    def motor_calibration(self):
        self._find_odrive()
        print("Calibrating Odrive for brushless motor (you should hear a "
        "beep)...")

        #self.odrv_axis.requested_state = AXIS_STATE_MOTOR_CALIBRATION (..gave error)
        self.odrv_axis.requested_state = AXIS_STATE_FULL_CALIBRATION_SEQUENCE


        # Wait for calibration to take place
        while self.odrv_axis.current_state != AXIS_STATE_IDLE:
            time.sleep(0.5)

        if self.odrv_axis.motor.error != 0:
            print("Error: Odrive reported an error of {} while in the state "
            "AXIS_STATE_FULL_CALIBRATION_SEQUENCE. Printing out Odrive motor data for "
            "debug:\n{}".format(self.odrv_axis.motor.error,
                                self.odrv_axis.motor))

            sys.exit(1)

        if self.odrv_axis.motor.config.phase_inductance <= self.MIN_PHASE_INDUCTANCE or \
        self.odrv_axis.motor.config.phase_inductance >= self.MAX_PHASE_INDUCTANCE:
            print("Error: After odrive motor calibration, the phase inductance "
            "is at {}, which is outside of the expected range. Either widen the "
            "boundaries of MIN_PHASE_INDUCTANCE and MAX_PHASE_INDUCTANCE (which "
            "is between {} and {} respectively) or debug/fix your setup. Printing "
            "out Odrive motor data for debug:\n{}".format(self.odrv_axis.motor.config.phase_inductance,
                                                          self.MIN_PHASE_INDUCTANCE,
                                                          self.MAX_PHASE_INDUCTANCE,
                                                          self.odrv_axis.motor))

            sys.exit(1)

        if self.odrv_axis.motor.config.phase_resistance <= self.MIN_PHASE_RESISTANCE or \
        self.odrv_axis.motor.config.phase_resistance >= self.MAX_PHASE_RESISTANCE:
            print("Error: After odrive motor calibration, the phase resistance "
            "is at {}, which is outside of the expected range. Either raise the "
            "MAX_PHASE_RESISTANCE (which is between {} and {} respectively) or "
            "debug/fix your setup. Printing out Odrive motor data for "
            "debug:\n{}".format(self.odrv_axis.motor.config.phase_resistance,
                                self.MIN_PHASE_RESISTANCE,
                                self.MAX_PHASE_RESISTANCE,
                                self.odrv_axis.motor))

            sys.exit(1)

        # If all looks good, then saving the calibration to persistent memory is OK
        self.odrv_axis.motor.config.pre_calibrated = True

        print("Saving motor calibration configuration and rebooting...")
        try:
            self.odrv.save_configuration() # includes reboot
            print("Motor calibration configuration saved.")

        except fibre.libfibre.ObjectLostError:
            print("Device disconnected")
            pass # Saving configuration makes the device reboot

    def encoder_calibration(self):
        self._find_odrive()

        # When the ..config.use_index = True, the index must be found before offset
        # calibration is run
        print("Encoder calibration. Searching for encoder index...")
        self.odrv_axis.requested_state = AXIS_STATE_ENCODER_INDEX_SEARCH

        # Check the alignment between the motor and the encoder sensor.
        print("Calibrating encoder offset...")
        self.odrv_axis.requested_state = AXIS_STATE_ENCODER_OFFSET_CALIBRATION

        # Wait for calibration to take place
        while self.odrv_axis.current_state != AXIS_STATE_IDLE:
            time.sleep(0.5)

        if self.odrv_axis.encoder.error != 0:
            print("Error: Odrive reported an error of {} while in the state "
            "AXIS_STATE_ENCODER_OFFSET_CALIBRATION. Printing out Odrive encoder "
            "data for debug:\n{}".format(self.odrv_axis.encoder.error,
                                         self.odrv_axis.encoder))

            sys.exit(1)

        # TODO: Evaluate need for this check.
        '''
        # If offset_float isn't 0.5 within some tolerance, or its not 1.5 within
        # some tolerance, raise an error
        if not ((self.odrv_axis.encoder.config.offset_float > 0.5 - self.ENCODER_OFFSET_FLOAT_TOLERANCE and \
        self.odrv_axis.encoder.config.offset_float < 0.5 + self.ENCODER_OFFSET_FLOAT_TOLERANCE) or \
        (self.odrv_axis.encoder.config.offset_float > 1.5 - self.ENCODER_OFFSET_FLOAT_TOLERANCE and \
        self.odrv_axis.encoder.config.offset_float < 1.5 + self.ENCODER_OFFSET_FLOAT_TOLERANCE)):
            print("Error: After odrive encoder calibration, the 'offset_float' "
            "is at {}, which is outside of the expected range. 'offset_float' "
            "should be close to 0.5 or 1.5 with a tolerance of {}. Either "
            "increase the tolerance or debug/fix your setup. Printing out "
            "Odrive encoder data for debug:\n{}".format(self.odrv_axis.encoder.config.offset_float,
                                                        self.ENCODER_OFFSET_FLOAT_TOLERANCE,
                                                        self.odrv_axis.encoder))

            sys.exit(1)
        '''

        # If all looks good, then lets tell ODrive that saving this calibration
        # to persistent memory is OK
        self.odrv_axis.encoder.config.pre_calibrated = True

    def configure(self):
        """
        Configures the odrive device for brushless motor.
        """
        self.set_odrive_parameters()
        # input("Make sure the motor is free to move, then press enter...")
        self.motor_calibration()
        print("Odrive basic configuration and motor calibration finished.")

    def full_calibration(self):
        """
        Make full calibration (MOTOR + ENCODER) of the motor.
        """

        print("Full calibration ")
        self.odrv_axis.requested_state = AXIS_STATE_FULL_CALIBRATION_SEQUENCE
        
        # Wait for calibration to take place
        while self.odrv_axis.current_state != AXIS_STATE_IDLE:
            time.sleep(0.5)

    def mode_idle(self):
        """
        Puts the motor in idle (i.e. can move freely).
        """

        self.odrv_axis.requested_state = AXIS_STATE_IDLE

    def mode_close_loop_control(self):
        """
        Puts the motor in closed loop control.
        """
        self.odrv_axis.requested_state = AXIS_STATE_CLOSED_LOOP_CONTROL
        
        time.sleep(0.5)
        
        if self.odrv_axis.current_state!= AXIS_STATE_CLOSED_LOOP_CONTROL:
            print("Failed to enter closed loop control, check motor and encoder ")
            sys.exit(1)
        print("Controller for current axis set to closed loop control")

    def move_input_pos(self, angle):
        """
        Puts the motor at a certain angle.

        :param angle: Angle you want the motor to move.
        :type angle: int or float
        """
        # TODO:
        # If the shadow count of the encoder does not change..the motor has not
        # moved and there is an error somewhere. We need to raise an exception + print
        # debug info.
        init_pos = self.odrv_axis.encoder.shadow_count
        current_pos = init_pos

        # Give move command..
        self.odrv_axis.controller.input_pos = angle/360.0
        time.sleep(0.5)

        current_pos = self.odrv_axis.encoder.shadow_count
        if current_pos > init_pos:
            print("Test passed!")
            return
        print("Error: Failed to move motor, check motor and encoder errors.")
        sys.exit(1)


if __name__ == "__main__":
    motor_axis0_config = BrushlessMotorConfig(axis_num = 0)
    #motor_axis1_config = BrushlessMotorConfig(axis_num = 1)
    motor_axis0_config.erase_config()

    # Configure basic parameters for Odrive and calibrate motor
    motor_axis0_config.configure()
    #motor_axis1_config.configure()

    # Perform encoder calibration
    motor_axis0_config.encoder_calibration()
    #motor_axis1_config.encoder_calibration()
   
    print("CONDUCTING MOTOR TESTS FOR AXIS 0")
    motor_axis0_config.mode_close_loop_control()

    print("Continuing test in 5 seconds..")
    time.sleep(5)

    # Go from 0 to 360 degrees in increments of 30 degrees
    for angle in range(0, 390, 30):
        print("Setting motor to {} degrees.".format(angle))
        motor_axis0_config.move_input_pos(angle)
        time.sleep(2)

    print("Placing motor for axis 0 in idle. If you move motor, motor will "
          "move freely")
    motor_axis0_config.mode_idle()

    print("Continuing test in 5 seconds..")
    time.sleep(5)
    
    '''print("CONDUCTING MOTOR TESTS FOR AXIS 1")
    motor_axis1_config.mode_close_loop_control()

    # Go from 0 to 360 degrees in increments of 30 degrees
    for angle in range(0, 390, 30):
        print("Setting motor to {} degrees.".format(angle))
        motor_axis1_config.move_input_pos(-angle)
        time.sleep(2)

    print("Placing motor for axis 1 in idle. If you move motor, motor will "
          "move freely")
    motor_axis1_config.mode_idle()
    '''